@@ -115,40 +115,30 @@ RsvgPaintServer *
 rsvg_paint_server_parse (gboolean * inherit, const RsvgDefs *defs, const char *str,
 						 guint32 current_color)
 {
+	char *name;
 	guint32 rgb;
 	if (inherit != NULL)
 		*inherit = 1;
 	if (!strcmp (str, "none"))
 		return NULL;
 
-	if (!strncmp (str, "url(", 4))
+	name = rsvg_get_url_string (str);
+	if (name)
 		{
-			const char *p = str + 4;
-			int ix;
-			char *name;
 			RsvgNode *val;
-			
-			while (g_ascii_isspace (*p)) p++;
-			for (ix = 0; p[ix]; ix++)
-				if (p[ix] == ')') break;
-			if (p[ix] != ')')
-				return NULL;
-			name = g_strndup (p, ix);
 			val = rsvg_defs_lookup (defs, name);
 			g_free (name);
+
 			if (val == NULL)
 				return NULL;
-			switch (val->type)
-				{
-				case RSVG_NODE_LINGRAD:
-					return rsvg_paint_server_lin_grad ((RsvgLinearGradient *)val);
-				case RSVG_NODE_RADGRAD:
-					return rsvg_paint_server_rad_grad ((RsvgRadialGradient *)val);
-				case RSVG_NODE_PATTERN:
-					return rsvg_paint_server_pattern ((RsvgPattern *)val);
-				default:
-					return NULL;
-				}
+			if (!strcmp(val->type->str, "linearGradient"))
+				return rsvg_paint_server_lin_grad ((RsvgLinearGradient *)val);
+			else if (!strcmp(val->type->str, "radialGradient"))
+				return rsvg_paint_server_rad_grad ((RsvgRadialGradient *)val);
+			else if (!strcmp(val->type->str, "pattern"))
+				return rsvg_paint_server_pattern ((RsvgPattern *)val);
+			else
+				return NULL;
 		}
 	else if (!strcmp (str, "inherit"))
 		{
@@ -206,6 +196,7 @@ rsvg_stop_set_atts (RsvgNode *self, RsvgHandle *ctx,
 	gboolean is_current_color = FALSE;
 	const char *value;
 	RsvgGradientStop * stop;
+	RsvgState state;
 
 	stop = (RsvgGradientStop *)self;
 
@@ -214,7 +205,8 @@ rsvg_stop_set_atts (RsvgNode *self, RsvgHandle *ctx,
 			if ((value = rsvg_property_bag_lookup (atts, "offset")))
 				{
 					/* either a number [0,1] or a percentage */
-					offset = rsvg_css_parse_normalized_length (value, rsvg_dpi_percentage (ctx), 1., 0.);
+					RsvgLength length = _rsvg_css_parse_length (value);
+					offset = _rsvg_css_hand_normalize_length(&length, rsvg_dpi_percentage (ctx), 1., 0.);
 					
 					if (offset < 0.)
 						offset = 0.;
@@ -231,8 +223,7 @@ rsvg_stop_set_atts (RsvgNode *self, RsvgHandle *ctx,
 
 			rsvg_parse_style_pairs (ctx, self->state, atts);
 		}
-	self->parent = ctx->currentnode;
-	RsvgState state;
+	self->parent = ctx->priv->currentnode;
 	rsvg_state_init(&state);
 	rsvg_state_reconstruct(&state, self);
 	if (is_current_color)
@@ -249,7 +240,6 @@ rsvg_new_stop (void)
 	stop->super.set_atts = rsvg_stop_set_atts;
 	stop->offset = 0;
 	stop->rgba = 0;
-	stop->super.type = RSVG_NODE_STOP;
 	return &stop->super;
 }
 
@@ -258,26 +248,25 @@ rsvg_linear_gradient_set_atts (RsvgNode * self, RsvgHandle *ctx, RsvgPropertyBag
 {
 	RsvgLinearGradient *grad = (RsvgLinearGradient *)self;
 	const char *value;
-	double font_size = rsvg_state_current_font_size(ctx);
 		
 	if (rsvg_property_bag_size (atts))
 		{
 			if ((value = rsvg_property_bag_lookup (atts, "id")))
-				rsvg_defs_register_name (ctx->defs, value, self);
+				rsvg_defs_register_name (ctx->priv->defs, value, self);
 			if ((value = rsvg_property_bag_lookup (atts, "x1"))){
-				grad->x1 = rsvg_css_parse_normalized_length (value, ctx->dpi_x, 1, font_size);
+				grad->x1 = _rsvg_css_parse_length (value);
 				grad->hasx1 = TRUE;
 			}				
 			if ((value = rsvg_property_bag_lookup (atts, "y1"))){
-				grad->y1 = rsvg_css_parse_normalized_length (value, ctx->dpi_y, 1, font_size);
+				grad->y1 = _rsvg_css_parse_length (value);
 				grad->hasy1 = TRUE;
 			}				
 			if ((value = rsvg_property_bag_lookup (atts, "x2"))){
-				grad->x2 = rsvg_css_parse_normalized_length (value, ctx->dpi_x, 1, font_size);
+				grad->x2 = _rsvg_css_parse_length (value);
 				grad->hasx2 = TRUE;
 			}	
 			if ((value = rsvg_property_bag_lookup (atts, "y2"))){
-				grad->y2 = rsvg_css_parse_normalized_length (value, ctx->dpi_y, 1, font_size);
+				grad->y2 = _rsvg_css_parse_length (value);
 				grad->hasy2 = TRUE;
 			}	
 		if ((value = rsvg_property_bag_lookup (atts, "spreadMethod")))
@@ -294,7 +283,7 @@ rsvg_linear_gradient_set_atts (RsvgNode * self, RsvgHandle *ctx, RsvgPropertyBag
 					grad->hasspread = TRUE;
 				}
 			if ((value = rsvg_property_bag_lookup (atts, "xlink:href")))
-				rsvg_defs_add_resolver (ctx->defs, &grad->fallback, value);
+				rsvg_defs_add_resolver (ctx->priv->defs, &grad->fallback, value);
 			if ((value = rsvg_property_bag_lookup (atts, "gradientTransform"))){
 				rsvg_parse_transform (grad->affine, value);
 				grad->hastransform = TRUE;
@@ -320,13 +309,10 @@ rsvg_new_linear_gradient (void)
 	RsvgLinearGradient *grad = NULL;
 	grad = g_new (RsvgLinearGradient, 1);
 	_rsvg_node_init(&grad->super);
-	grad->super.type = RSVG_NODE_LINGRAD;
 	_rsvg_affine_identity(grad->affine);
 	grad->has_current_color = FALSE;
-	grad->x1 = 0;
-	grad->y1 = 0;
-	grad->x2 = 1.0;
-	grad->y2 = 0;
+	grad->x1 = grad->y1 = grad->y2 = _rsvg_css_parse_length ("0");
+	grad->x2 = _rsvg_css_parse_length ("1");
 	grad->fallback = NULL;
 	grad->obj_bbox = TRUE;
 	grad->spread = RSVG_GRADIENT_PAD;
@@ -340,38 +326,37 @@ rsvg_radial_gradient_set_atts (RsvgNode * self, RsvgHandle *ctx, RsvgPropertyBag
 {
 	RsvgRadialGradient *grad = (RsvgRadialGradient *) self;
 	const char *value;
-	double font_size = rsvg_state_current_font_size(ctx);
 	
 	if (rsvg_property_bag_size (atts))
 		{
 			if ((value = rsvg_property_bag_lookup (atts, "id")))
-				rsvg_defs_register_name (ctx->defs, value, self);
+				rsvg_defs_register_name (ctx->priv->defs, value, self);
 			if ((value = rsvg_property_bag_lookup (atts, "cx"))){
-				grad->cx = rsvg_css_parse_normalized_length (value, ctx->dpi_x, 1, font_size);	
+				grad->cx = _rsvg_css_parse_length (value);
 				grad->hascx = TRUE;
 				if (!grad->hasfx)
 					grad->fx = grad->cx;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "cy"))){
-				grad->cy = rsvg_css_parse_normalized_length (value, ctx->dpi_y, 1, font_size);	
+				grad->cy = _rsvg_css_parse_length (value);	
 				grad->hascy = TRUE;
 				if (!grad->hasfy)
 					grad->fy = grad->cy;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "r"))){
-				grad->r = rsvg_css_parse_normalized_length (value, rsvg_dpi_percentage (ctx), 1, font_size);	
+				grad->r = _rsvg_css_parse_length (value);	
 				grad->hasr = TRUE;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "fx"))){
-				grad->fx = rsvg_css_parse_normalized_length (value, ctx->dpi_x, 1, font_size);	
+				grad->fx = _rsvg_css_parse_length (value);	
 				grad->hasfx = TRUE;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "fy"))){
-				grad->fy = rsvg_css_parse_normalized_length (value, ctx->dpi_y, 1, font_size);
+				grad->fy = _rsvg_css_parse_length (value);
 				grad->hasfy = TRUE;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "xlink:href")))
-				rsvg_defs_add_resolver (ctx->defs, &grad->fallback, value);	
+				rsvg_defs_add_resolver (ctx->priv->defs, &grad->fallback, value);	
 			if ((value = rsvg_property_bag_lookup (atts, "gradientTransform"))){
 				rsvg_parse_transform (grad->affine, value);
 				grad->hastransform = TRUE;
@@ -406,17 +391,12 @@ rsvg_new_radial_gradient (void)
 
 	RsvgRadialGradient * grad = g_new (RsvgRadialGradient, 1);
 	_rsvg_node_init(&grad->super);
-	grad->super.type = RSVG_NODE_RADGRAD;
 	_rsvg_affine_identity(grad->affine);
 	grad->has_current_color = FALSE;
 	grad->obj_bbox = TRUE;
 	grad->spread = RSVG_GRADIENT_PAD;
 	grad->fallback = NULL;
-	grad->cx = 0.5;
-	grad->cy = 0.5;
-	grad->r = 0.5;
-	grad->fx = 0.5;
-	grad->fy = 0.5;
+	grad->cx = grad->cy = grad->r = grad->fx = grad->fy = _rsvg_css_parse_length("0.5");
 	grad->super.set_atts = rsvg_radial_gradient_set_atts;
 	grad->hascx = grad->hascy = grad->hasfx = grad->hasfy = grad->hasr = grad->hasbbox = grad->hasspread = grad->hastransform = FALSE;
 	return &grad->super;
@@ -427,38 +407,38 @@ rsvg_pattern_set_atts (RsvgNode * self, RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
 	RsvgPattern *pattern = (RsvgPattern *)self;
 	const char  *value;
-	double font_size = rsvg_state_current_font_size(ctx);
 
 	if (rsvg_property_bag_size (atts))
 		{
 			if ((value = rsvg_property_bag_lookup (atts, "id")))
-				rsvg_defs_register_name (ctx->defs, value, self);
+				rsvg_defs_register_name (ctx->priv->defs, value, self);
 			if ((value = rsvg_property_bag_lookup (atts, "viewBox"))){
-				pattern->vbox = rsvg_css_parse_vbox (value, 
-													 &pattern->vbx, 
-													 &pattern->vby,
-													 &pattern->vbw, 
-													 &pattern->vbh);
+				pattern->vbox = rsvg_css_parse_vbox (value);
 				pattern->hasvbox = TRUE;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "x"))){
-				pattern->x = rsvg_css_parse_normalized_length (value, ctx->dpi_x, 1, font_size);
+				pattern->x = _rsvg_css_parse_length (value);
 				pattern->hasx = TRUE;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "y"))){
-				pattern->y = rsvg_css_parse_normalized_length (value, ctx->dpi_y, 1, font_size);
+				pattern->y = _rsvg_css_parse_length (value);
 				pattern->hasy = TRUE;			
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "width"))){
-				pattern->width = rsvg_css_parse_normalized_length (value, ctx->dpi_x, 1, font_size);
+				pattern->width = _rsvg_css_parse_length (value);
 				pattern->haswidth = TRUE;
 			}
 			if ((value = rsvg_property_bag_lookup (atts, "height"))){
-				pattern->height = rsvg_css_parse_normalized_length (value, ctx->dpi_y, 1, font_size);
+				pattern->height = _rsvg_css_parse_length (value);
 				pattern->hasheight = TRUE;
 			}
-			if ((value = rsvg_property_bag_lookup (atts, "xlink:href")))
-				rsvg_defs_add_resolver (ctx->defs, (RsvgNode **)&pattern->fallback, value);
+			if ((value = rsvg_property_bag_lookup (atts, "xlink:href"))) {
+				/* The (void *) cast is to avoid a GCC warning like:
+				 * "warning: dereferencing type-punned pointer will break strict-aliasing rules"
+				 * which is wrong for this code. (void *) introduces a compatible
+				 * intermediate type in the cast list. */
+				 rsvg_defs_add_resolver (ctx->priv->defs, (RsvgNode **)(void *)&pattern->fallback, value);
+			}
 			if ((value = rsvg_property_bag_lookup (atts, "patternTransform"))){
 				rsvg_parse_transform (pattern->affine, value);
 				pattern->hastransform = TRUE;
@@ -492,21 +472,14 @@ rsvg_new_pattern (void)
 	_rsvg_node_init(&pattern->super);
 	pattern->obj_bbox = TRUE;
 	pattern->obj_cbbox = FALSE;
-	pattern->x = 0;
-	pattern->y = 0;
-	pattern->width = 0;
-	pattern->height = 0;
-	pattern->vbx = 0;
-	pattern->vby = 0;
-	pattern->vbw = 1;
-	pattern->vbh = 1;
+	pattern->x = pattern->y = pattern->width = pattern->height = 
+		_rsvg_css_parse_length("0");
 	pattern->fallback = NULL;
 	pattern->preserve_aspect_ratio = RSVG_ASPECT_RATIO_XMID_YMID;
-	pattern->vbox = FALSE;
+	pattern->vbox.active = FALSE;
 	_rsvg_affine_identity(pattern->affine);
 	pattern->super.set_atts = rsvg_pattern_set_atts;
-	pattern->super.type = RSVG_NODE_PATTERN;
-	pattern->hasx = pattern->hasy = pattern->haswidth = pattern->hasheight = pattern->hasvbox = pattern->hasbbox = pattern->hascbox = pattern->hasaspect = pattern->hastransform = pattern->hasaspect = FALSE; 
+	pattern->hasx = pattern->hasy = pattern->haswidth = pattern->hasheight = pattern->hasbbox = pattern->hascbox = pattern->hasvbox = pattern->hasaspect = pattern->hastransform = pattern->hasaspect = FALSE; 
 	return &pattern->super;
 }
 
@@ -515,7 +488,7 @@ static int hasstop(GPtrArray *lookin)
 	unsigned int i;
 	for (i = 0; i < lookin->len; i++)
 		{
-			if (((RsvgNode *)g_ptr_array_index(lookin, i))->type == RSVG_NODE_STOP)
+			if (!strcmp(((RsvgNode *)g_ptr_array_index(lookin, i))->type->str, "stop"))
 				return 1;
 		}
 	return 0;
@@ -529,7 +502,7 @@ rsvg_linear_gradient_fix_fallback(RsvgLinearGradient * grad)
 	ufallback = grad->fallback;
 	while (ufallback != NULL)
 		{
-			if (ufallback->type == RSVG_NODE_LINGRAD){
+			if (!strcmp(ufallback->type->str, "linearGradient")){
 				RsvgLinearGradient * fallback = (RsvgLinearGradient *)ufallback;
 				if (!grad->hasx1 && fallback->hasx1){
 					grad->hasx1 = TRUE;
@@ -565,7 +538,7 @@ rsvg_linear_gradient_fix_fallback(RsvgLinearGradient * grad)
 				}
 				ufallback = fallback->fallback; 
 			}
-			else if (ufallback->type == RSVG_NODE_RADGRAD){
+			else if (!strcmp(ufallback->type->str, "radialGradient")){
 				RsvgRadialGradient * fallback = (RsvgRadialGradient *)ufallback;
 				if (!grad->hastransform && fallback->hastransform){
 					grad->hastransform = TRUE;
@@ -596,7 +569,7 @@ rsvg_radial_gradient_fix_fallback(RsvgRadialGradient * grad)
 	ufallback = grad->fallback;
 	while (ufallback != NULL)
 		{
-			if (ufallback->type == RSVG_NODE_RADGRAD){
+			if (!strcmp(ufallback->type->str, "radialGradient")){
 				RsvgRadialGradient * fallback = (RsvgRadialGradient *)ufallback;
 				if (!grad->hascx && fallback->hascx){
 					grad->hascx = TRUE;
@@ -636,7 +609,7 @@ rsvg_radial_gradient_fix_fallback(RsvgRadialGradient * grad)
 				}
 				ufallback = fallback->fallback; 
 			}
-			else if (ufallback->type == RSVG_NODE_LINGRAD){
+			else if (!strcmp(ufallback->type->str, "linearGradient")){
 				RsvgLinearGradient * fallback = (RsvgLinearGradient *)ufallback;
 				if (!grad->hastransform && fallback->hastransform){
 					grad->hastransform = TRUE;
@@ -689,11 +662,6 @@ rsvg_pattern_fix_fallback(RsvgPattern * pattern)
 						pattern->affine[i] = fallback->affine[i];
 				}
 			if (!pattern->hasvbox && fallback->hasvbox){
-				pattern->hasvbox = TRUE;
-				pattern->vbx = fallback->vbx;
-				pattern->vby = fallback->vby;
-				pattern->vbw = fallback->vbw;
-				pattern->vbh = fallback->vbh;
 				pattern->vbox = fallback->vbox;
 			}
 			if (!pattern->hasaspect && fallback->hasaspect){
